(* file & console logging *)
(* solyga@gmx.de 2021-01-13 *)
IMPLEMENTATION MODULE Log;

IMPORT IO, FIO, FIOStd, Str;
IMPORT WinBase;
IMPORT Types, SO, Lib2, Lib3, Cfg, Mtxs, Tmr, Utils;

CONST
  mid = Cfg.midLog;

  lenMax = 128 * 1024 * 1024;  (* max file len *)
  cntMax = 16;       (* historic logs *)
  clean = Cfg.cln;

  lenMaxLine = 512;  (* max chars per line *)

  fnmBase = Cfg.prg;
  wdFileNums = 3;
  fnmExt  = '.log';
  fnmLen = SIZE( fnmBase ) - 1 + wdFileNums + SIZE( fnmExt ) - 1;

  wdModName = 8;

  noFile ::= FIOStd.inv;

TYPE
  TFileIdx = [ 0 .. cntMax ];
  TFileName = ARRAY [ 1 .. fnmLen ] OF CHAR;

  TModDesc = RECORD
    n: TModName;
    r: BOOLEAN;
  END;

TYPE
  TLvlStrs = ARRAY TLvl OF CHAR;

CONST
  lvlStrs = TLvlStrs{ 'M', 'W', 'E', 'F' };

VAR
  f: FIO.File;
  mods: ARRAY Cfg.TModId OF TModDesc;
  fnms: ARRAY TFileIdx OF TFileName;
  mtx: Mtxs.TMtx;


PROCEDURE Enter();
BEGIN
  Mtxs.Req( mtx );
END Enter;

PROCEDURE Leave();
BEGIN
  Mtxs.Rel( mtx );
END Leave;


PROCEDURE MkFileName( VAR n: TFileName; i: TFileIdx );
VAR
  b: SO.TBuffer;
  s: ARRAY [1..SIZE(n)] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, fnmBase );
  IF i > 0 THEN
    b.c.p := '0';
    SO.WrCard( b, i, wdFileNums );
  END;
  SO.WrStr( b, fnmExt );
  Str.Copy( n, s );
END MkFileName;


PROCEDURE WrTicks( VAR b: SO.TBuffer );
VAR
  t: Types.TLngCard;
  c: CHAR;
BEGIN
  t := Tmr.Ticks();
  c := b.c.p;
  b.c.p := '0';
  SO.WrLngHex( b, t, 8 );
  SO.WrChar( b, ' ' );
  b.c.p := c;
END WrTicks;

PROCEDURE WrTime( VAR b: SO.TBuffer );
VAR
  t: WinBase.SYSTEMTIME;
  c: CHAR;
BEGIN
  WinBase.GetSystemTime( t );
  (*
  WinBase.GetLocalTime( t );
  *)
  c := b.c.p;
  b.c.p := '0';
  SO.WrCard( b, t.wYear, 4 );
  SO.WrCard( b, t.wMonth, 2 );
  SO.WrCard( b, t.wDay, 2 );
  SO.WrChar( b, ' ' );
  SO.WrCard( b, t.wHour, 2 );
  SO.WrCard( b, t.wMinute, 2 );
  SO.WrCard( b, t.wSecond, 2 );
  SO.WrChar( b, '.' );
  SO.WrCard( b, t.wMilliseconds, 3 );
  SO.WrChar( b, ' ' );
  b.c.p := c;
END WrTime;

PROCEDURE WrModName( VAR b: SO.TBuffer; m: Cfg.TModId );
BEGIN
  SO.WrStrAdj( b, mods[ m ].n, -(wdModName+1) );
END WrModName;


PROCEDURE Open();
BEGIN
  IF ~ Cfg.log THEN RETURN END;
  IF FIO.Exists( fnms[0] ) THEN
    f := FIO.Append( fnms[0] );
  ELSE
    f := FIO.Create( fnms[0] );
  END;
END Open;

PROCEDURE Close();
BEGIN
  IF f # noFile THEN
    FIO.Close( f );
    f := noFile;
  END;
END Close;

PROCEDURE Size(): Types.TLngCard;
BEGIN
  IF f = noFile THEN RETURN 0 END;
  RETURN FIO.Size( f );
END Size;


PROCEDURE Rotate();
VAR
  i: TFileIdx;
BEGIN
  IF ~ Cfg.log THEN RETURN END;  (* don't even touch ancient log files *)
  Close();
  IF FIO.Exists( fnms[cntMax] ) THEN FIO.Erase( fnms[cntMax] ) END;
  FOR i := cntMax - 1 TO 0 BY -1 DO
    IF FIO.Exists( fnms[i] ) THEN FIO.Rename( fnms[i], fnms[i+1] ) END;
  END;
  Open();
END Rotate;


PROCEDURE Write( m: Cfg.TModId; l: TLvl; x-: ARRAY OF CHAR; screen: BOOLEAN );
VAR
  b: SO.TBuffer;
  s: ARRAY [1..lenMaxLine] OF CHAR;
BEGIN
  Enter();  (* mods, I/O *)
  SO.Init( b, s );
  WrTime( b );
  WrTicks( b );
  WrModName( b, m );
  SO.WrStr( b, lvlStrs[ l ] );
  SO.WrChar( b, ' ' );
  SO.WrStr( b, x );
  SO.WrLn( b );
  IF Size() + b.l > lenMax THEN Rotate() END;
  IF f # noFile THEN FIO.WrStr( f, s ) END;
  IF screen THEN
    SO.Init( b, s );
    SO.WrStr( b, lvlStrs[ l ] );
    SO.WrChar( b, ' ' );
    SO.WrStr( b, mods[ m ].n );
    SO.WrChar( b, ' ' );
    SO.WrStr( b, x );
    IO.WrStr( s );
    IO.WrLn();
  END;
  Leave();
END Write;

PROCEDURE Register( m: Cfg.TModId; t-: ARRAY OF CHAR );
VAR
  n: TModName;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..lenMaxLine] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Register ' );
    SO.WrCard( b, ORD( m ), 2 );
    SO.WrStr( b, ' "' );
    SO.WrStr( b, n );
    SO.WrChar( b, '"' );
    Write( mid, lvlMsg, s, FALSE );
  END Show;

  PROCEDURE Fatal( s-: ARRAY OF CHAR );
  BEGIN
    Write( mid, lvlFat, s, TRUE );
    Lib3.Halt();
  END Fatal;

VAR
  l: CARDINAL;
  b: SO.TBuffer;
  s: ARRAY [1..lenMaxLine] OF CHAR;
  i: Cfg.TModId;
BEGIN
  Enter();  (* mods *)
  l := Str.Length( t );
  IF ( l = 0 ) OR ( l > SIZE( n ) ) THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Register Invalid name "' );
    SO.WrStr( b, t );
    SO.WrChar( b, '"' );
    Fatal( s );
  END;
  Str.Copy( n, t );
  IF mods[ m ].r THEN
    SO.Init( b, s );
    SO.WrStr( b, 'Register Id known ' );
    SO.WrCard( b, ORD( m ), 2 );
    SO.WrStr( b, ' "' );
    SO.WrStr( b, mods[ m ].n );
    SO.WrChar( b, '"' );
    Fatal( s );
  END;
  FOR i := MIN( Cfg.TModId ) TO MAX( Cfg.TModId ) DO
    IF mods[ i ].r & ( Str.Compare( mods[ i ].n, n ) = 0 ) THEN
      SO.Init( b, s );
      SO.WrStr( b, 'Register Name known ' );
      SO.WrCard( b, ORD( i ), 2 );
      SO.WrStr( b, ' "' );
      SO.WrStr( b, n );
      SO.WrChar( b, '"' );
      Fatal( s );
    END;
  END;
  mods[ m ].n := n;
  mods[ m ].r := TRUE;
  Show();
  Leave();
END Register;


VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
VAR
  b: SO.TBuffer;
  s: ARRAY [1..64] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'DoneModule rc ' );
  SO.WrCard( b, Lib3.ReturnCode(), 1 );
  Write( mid, lvlMsg, s, FALSE );
  Close();
  oldTerm();
END DoneModule;

PROCEDURE InitModule();

  PROCEDURE InitMods();
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..SIZE(TModName)] OF CHAR;
    i: Cfg.TModId;
  BEGIN
    SO.Init( b, s );
    b.c.p := '0';
    FOR i := MIN( Cfg.TModId ) TO MAX( Cfg.TModId ) DO
      SO.Clear( b );
      SO.WrHex( b, ORD( i ), wdModName );
      Str.Copy( mods[ i ].n, s );
    END;
  END InitMods;

  PROCEDURE InitNames();
  VAR
    i: TFileIdx;
  BEGIN
    FOR i := MIN( TFileIdx ) TO MAX( TFileIdx ) DO
      MkFileName( fnms[ i ], i );
    END;
  END InitNames;

  PROCEDURE Clean();
  VAR
    i: TFileIdx;
  BEGIN
    IF ~ Cfg.log OR ~ Cfg.cln THEN RETURN END;
    FOR i := MIN( TFileIdx ) TO MAX( TFileIdx ) DO
      IF FIO.Exists( fnms[i] ) THEN FIO.Erase( fnms[i] ) END;
    END;
  END Clean;

  PROCEDURE Mark();
  BEGIN
    Write( Cfg.midNull, lvlMsg, '========================================', FALSE );
  END Mark;

  PROCEDURE Ticks();
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..16] OF CHAR;
  BEGIN
    SO.Init( b, s );
    SO.WrStr( b, 'ticks ' );
    SO.WrLngHex( b, Tmr.Ticks(), 8 );
    Write( mid, lvlMsg, s, FALSE );
  END Ticks;

BEGIN
  f := noFile;
  Lib2.Fill( mods, 0 );
  Lib2.Fill( fnms, 0 );
  mtx := Mtxs.New();

  InitMods();
  InitNames();
  Clean();
  Open();
  Mark();
  Register( mid, 'Log' );
  Ticks();

  Lib3.Terminate( DoneModule, oldTerm );
END InitModule;

BEGIN
  InitModule();
END Log.
